#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
)

import networkx as nx

if TYPE_CHECKING:
    from abses.human import BaseHuman


class LinkContainer:
    """连接容器"""

    def __init__(self) -> None:
        self._bipartite: Dict[str, bool] = {}
        self._graphs: Dict[str, nx.Graph] = {}

    @property
    def links(self) -> Tuple[str]:
        """所有关联类型"""
        return tuple(self._graphs.keys())

    def _is_bipartite(self, node_1, node_2) -> bool:
        """是否是二分图"""
        return type(node_1) is not type(node_2)

    def _is_new_links_graph(self, link: str) -> bool:
        """是否是新的图"""
        return link not in self.links

    def _is_node(self, node: LinkNode, check: bool = True) -> bool:
        """是否是节点"""
        flag = isinstance(node, LinkNode)
        if check and not flag:
            raise TypeError(f"{node} is not a LinkNode.")
        return flag

    def _add_new_graph(
        self, link: str, graph_cls: Type, is_bipartite: bool
    ) -> None:
        """添加新图"""
        if not issubclass(graph_cls, nx.Graph):
            raise TypeError(
                f"Initializing a new link type '{link}', you must assign a valid graph type, instead of {graph_cls}."
            )
        self._bipartite[link] = is_bipartite
        self._graphs[link] = graph_cls()

    def _is_directed_graph(self, graph: Type | str) -> bool:
        if isinstance(graph, str):
            if self._is_new_links_graph(graph):
                raise KeyError(f"Link {graph} does not exist.")
            graph = type(self._graphs[graph])
            return self._is_directed_graph(graph)
        return graph in (nx.MultiDiGraph, nx.DiGraph)

    def add_link(
        self,
        node_1,
        node_2,
        link: str,
        graph_cls: Type = None,
        mutual: Optional[bool] = None,
    ) -> bool:
        """添加关联"""
        self._is_node(node_1, check=True)
        self._is_node(node_2, check=True)
        if self._is_new_links_graph(link):
            is_bipartite = self._is_bipartite(node_1, node_2)
            self._add_new_graph(link, graph_cls, is_bipartite)
        if self._is_directed_graph(graph_cls) and mutual is None:
            raise TypeError(
                f"{graph_cls} is Directed Graph, mutual must be set."
            )
        self.linking(node_1, node_2, link, mutual=mutual)

    def linking(
        self,
        node_1: LinkNode,
        node_2: LinkNode,
        link: str,
        mutual: bool = True,
    ) -> None:
        """建立关联"""
        graph = self.get_graph(link)
        if self._bipartite[link]:
            graph.add_node(node_1, bipartite=node_1.breed)
            graph.add_node(node_2, bipartite=node_2.breed)
        node_1.links.add(link)
        node_2.links.add(link)
        graph.add_edge(node_1, node_2)
        if mutual:
            graph.add_edge(node_2, node_1)

    def get_graph(self, link: str) -> nx.Graph:
        """获取图"""
        if self._is_new_links_graph(link=link):
            raise KeyError(f"Link {link} does not exist.")
        return self._graphs[link]

    def linked(
        self,
        link: str,
        node: LinkNode,
    ) -> Iterator[LinkNode]:
        """获取相关联的所有其它主体"""
        graph = self.get_graph(link)
        if isinstance(graph, nx.DiGraph):
            return graph.successors(node)
        return nx.all_neighbors(graph=graph, node=node)


class LinkNode:
    """节点类"""

    def __init__(self) -> None:
        self._container: LinkContainer = None
        self._links: Set[str] = set()

    @classmethod
    @property
    def breed(cls) -> str:
        """种类"""
        return cls.__name__

    @property
    def links(self) -> List[str]:
        """该主体所有连接类型"""
        return self._links

    @property
    def container(self) -> LinkContainer:
        """连接容器"""
        return self._container

    @container.setter
    def container(self, container: LinkContainer) -> None:
        if not isinstance(container, LinkContainer):
            raise TypeError(f"{container} is not a LinkContainer.")
        self._container = container

    def link_to(
        self,
        agent: LinkNode,
        link: Optional[str],
        graph_cls: Type = nx.Graph,
        **kwargs,
    ) -> None:
        """将行动者与其它行动者或地块建立连接"""
        self.container.add_link(
            node_1=self, node_2=agent, link=link, graph_cls=graph_cls, **kwargs
        )

    def linked(self, link: str) -> Iterator[LinkNode]:
        """获取相关联的所有其它主体，如果不在连接容器中，则返回空列表"""
        return (
            iter([])
            if link not in self.links
            else self.container.linked(link, self)
        )



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
from numbers import Number
from typing import TYPE_CHECKING, Any, List, Optional, Self

import geopandas
import mesa_geo as mg
import numpy as np
import pyproj
import rasterio as rio
import rioxarray
import xarray as xr
from mesa.space import Coordinate
from mesa_geo.raster_layers import Cell
from rasterio import mask
from shapely import Geometry

from abses.modules import CompositeModule, Module

from .actor import Actor
from .cells import PatchCell
from .errors import ABSESpyError
from .sequences import ActorsList
from .tools.func import norm_choice

if TYPE_CHECKING:
    from abses.main import MainModel

logger = logging.getLogger(__name__)
logger.info("Using rioxarray version: %s", rioxarray.__version__)

DEFAULT_WORLD = {
    "width": 10,
    "height": 10,
    "resolution": 1,
}
CRS = "epsg:4326"


class PatchModule(Module, mg.RasterLayer):
    # 基础的空间模块，继承这个类来创建一个子模块。看[这个教程](../features/architectural_elegance.md)来了解模型结构。这也是一个栅格图层，继承自 `mesa-geo` 的`RasterLayer`类，并可以放置主体。
    """
    The spatial sub-module base class.
    Inherit from this class to create a submodule. Look at [this tutorial](../features/architectural_elegance.md) to understand the model structure. This is also a raster layer, inherited from the 'mesa-geo.RasterLayer' class, and can place agents.

    Attributes:
        cell_properties:
            The accessible attributes of cells stored in this layer. All `PatchCell` methods which are decorated by the decorator `raster_attribute` should be appeared here.
        attributes:
            All accessible attributes from this layer.
        file:
            If the module is created by reading a raster dataset, save the file path.
        shape2d:
            Raster shape in 2D (heigh, width).
        shape3d:
            Raster shape in 3D (1, heigh, width).
    """

    def __init__(self, model, name=None, **kwargs):
        Module.__init__(self, model, name=name)
        mg.RasterLayer.__init__(self, **kwargs)
        for cell in self.array_cells.flatten():
            cell.layer = self
        self._file = None

    @property
    def cell_properties(self) -> set[str]:
        """The accessible attributes of cells stored in this layer. All `PatchCell` methods which are decorated by the decorator `raster_attribute` should be appeared here."""
        return self.cell_cls.__attribute_properties__()

    @property
    def attributes(self) -> set[str]:
        """All accessible attributes from this layer."""
        return self._attributes | self.cell_properties

    @property
    def file(self) -> str | None:
        """If the module is created by reading a raster dataset, save the file path."""
        return self._file

    @property
    def shape2d(self) -> Coordinate:
        """Raster shape in 2D (height, width).
        This is useful when working with 2d `numpy.array`.
        """
        return self.height, self.width

    @property
    def shape3d(self) -> Coordinate:
        """Raster shape in 3D (1, heigh, width).
        This is useful when working with `rasterio` band.
        """
        return 1, self.height, self.width

    @property
    def array_cells(self) -> np.ndarray:
        """Array type of the `PatchCell` stored in this module."""
        return np.flipud(np.array(self.cells).T)

    @property
    def coords(self) -> Coordinate:
        """Coordinate system of the raster data.
        This is useful when working with `xarray.DataArray`.
        """
        x_arr = np.linspace(
            self.total_bounds[0], self.total_bounds[2], self.width
        )
        y_arr = np.linspace(
            self.total_bounds[3], self.total_bounds[1], self.height
        )
        return {
            "y": y_arr,
            "x": x_arr,
        }

    @classmethod
    def from_resolution(
        cls,
        model: MainModel,
        name: str = None,
        shape: Coordinate = (10, 10),
        crs: Optional[pyproj.CRS | str] = CRS,
        resolution: Number = 1,
        cell_cls: type[PatchCell] = PatchCell,
    ) -> Self:
        """Create a layer from resolution.

        Parameters:
            model:
                ABSESpy Model that the new module belongs.
            name:
                Name of the new module. If None (by default), using lowercase of the '__class__.__name__'. E.g., class NewModule -> newmodule.
            shape:
                Array shape (height, width) of the new module. For example, if shape=(3, 5), it means the new module will store 15 cells.
            crs:
                Coordinate Reference Systems. If passing a string object, should be able to parsed by `pyproj`. By default, we use CRS = "epsg:4326".
            resolution:
                Spatial Resolution when creating the coordinates. By default 1, it means shape (3, 5) will generate coordinates like {y: [0, 1, 2], x: [0, 1, 2, 3, 4]}. Similar, when using resolution=0.1, it will be {y: [.0, .1, .2], x: [.0, .1, .2, .3, .4]}.
            cell_cls:
                Class type of `PatchCell` to create.

        Returns:
            A new instance of self ("PatchModule").
        """
        height, width = shape
        total_bounds = [0, 0, width * resolution, height * resolution]
        return cls(
            model,
            name=name,
            width=width,
            height=height,
            crs=crs,
            total_bounds=total_bounds,
            cell_cls=cell_cls,
        )

    @classmethod
    def copy_layer(
        cls,
        model: MainModel,
        layer: Self,
        name: Optional[str] = None,
        cell_cls: PatchCell = PatchCell,
    ) -> Self:
        """Copy an existing layer to create a new layer.

        Parameters:
            model:
                ABSESpy Model that the new module belongs.
            layer:
                Another layer to copy. These attributes will be copied: the coordinates, the crs, and the shape.
            name:
                Name of the new module. If None (by default), using lowercase of the '__class__.__name__'. E.g., class NewModule -> newmodule.
            cell_cls:
                Class type of `PatchCell` to create.

        Returns:
            A new instance of self ("PatchModule").
        """
        if not isinstance(layer, PatchModule):
            raise TypeError(f"{layer} is not a valid PatchModule.")

        return cls(
            model=model,
            name=name,
            width=layer.width,
            height=layer.height,
            crs=layer.crs,
            total_bounds=layer.total_bounds,
            cell_cls=cell_cls,
        )

    @classmethod
    def from_file(
        cls,
        raster_file: str,
        cell_cls: type[Cell] = PatchCell,
        attr_name: str | None = None,
        model: None | MainModel = None,
        name: str | None = None,
    ) -> Self:
        """Create a raster layer module from a file.

        Parameters:
            raster_file:
                File path of a geo-tiff dataset.
            model:
                ABSESpy Model that the new module belongs.
            attr_name:
                Assign a attribute name to the loaded raster data.
            name:
                Name of the new module. If None (by default), using lowercase of the '__class__.__name__'. E.g., class NewModule -> newmodule.
            cell_cls:
                Class type of `PatchCell` to create.

        """
        with rio.open(raster_file, "r") as dataset:
            values = dataset.read()
            _, height, width = values.shape
            total_bounds = [
                dataset.bounds.left,
                dataset.bounds.bottom,
                dataset.bounds.right,
                dataset.bounds.top,
            ]
        obj = cls(
            model=model,
            name=name,
            width=width,
            height=height,
            crs=dataset.crs,
            total_bounds=total_bounds,
            cell_cls=cell_cls,
        )
        obj._transform = dataset.transform
        obj._file = raster_file
        obj.apply_raster(values, attr_name=attr_name)
        return obj

    def _attr_or_array(self, data: None | str | np.ndarray) -> np.ndarray:
        """Determine the incoming data type and turn it into a reasonable array."""
        if data is None:
            return np.ones(self.shape2d)
        if isinstance(data, np.ndarray):
            if data.shape == self.shape2d:
                return data
            raise ABSESpyError(
                f"Shape mismatch: {data.shape} [input] != {self.shape2d} [expected]."
            )
        if isinstance(data, str) and data in self.attributes:
            return self.get_raster(data)
        raise TypeError("Invalid data type or shape.")

    def get_raster(self, attr_name: str | None = None) -> np.ndarray:
        """Obtaining the Raster layer by attribute.

        Parameters:
            attr_name:
                The attribute to retrieve. Update it if it is a dynamic variable. If None (by default), retrieve all attributes as a 3D array.

        Returns:
            A 3D array of attribute.
        """
        if attr_name not in self._dynamic_variables:
            return super().get_raster(attr_name)
        return self.dynamic_var(attr_name=attr_name).reshape(self.shape3d)

    def dynamic_var(self, attr_name: str) -> np.ndarray:
        """Update and get dynamic variable.

        Parameters:
            attr_name:
                The dynamic variable to retrieve.

        Returns:
            2D numpy.ndarray data of the variable.
        """
        array = super().dynamic_var(attr_name)
        # 判断算出来的是一个符合形状的矩阵
        self._attr_or_array(array)
        # 将矩阵转换为三维，并更新空间数据
        array_3d = array.reshape(self.shape3d)
        self.apply_raster(array_3d, attr_name=attr_name)
        return array

    def get_rasterio(self, attr_name: str | None = None) -> rio.MemoryFile:
        """Gets the Rasterio raster layer corresponding to the attribute. Save to a temporary rasterio memory file.

        Parameters:
            attr_name:
                The attribute name for creating the rasterio file.

        Returns:
            The rasterio tmp memory file of raster.
        """
        data = self.get_raster(attr_name=attr_name)
        # 如果获取到的是2维，重整为3维
        if len(data.shape):
            data = data.reshape(self.shape3d)
        with rio.MemoryFile() as mem_file:
            with mem_file.open(
                driver="GTiff",
                height=data.shape[1],
                width=data.shape[2],
                count=data.shape[0],  # number of bands
                dtype=str(data.dtype),
                crs=self.crs,
                transform=self.transform,
            ) as dataset:
                dataset.write(data)
            # Open the dataset again for reading and return
            return mem_file.open()

    def geometric_cells(self, geometry: Geometry, **kwargs) -> List[PatchCell]:
        """Gets all the cells that intersect the given geometry.

        Parameters:
            geometry:
                Shapely Geometry to search intersected cells.
            **kwargs:
                Args pass to the function `rasterio.mask.mask`. It influence how to build the mask for filtering cells. Please refer [this doc](https://rasterio.readthedocs.io/en/latest/api/rasterio.mask.html) for details.

        Returns:
            A list of PatchCell.
        """
        data = self.get_rasterio(list(self.attributes)[0])
        out_image, _ = mask.mask(data, [geometry], **kwargs)
        mask_ = out_image.reshape(self.shape2d)
        return list(self.array_cells[mask_.astype(bool)])

    def link_by_geometry(
        self, geo_agent: Actor, link: Optional[str] = None, **kwargs
    ) -> None:
        """Relates all cells intersecting a given geometry.

        Parameters:
            geo_agent:
                An Actor who has geometry info.
            link:
                The link info to save.
            **kwargs:
                Args pass to the function `rasterio.mask.mask`. It influence how to build the mask for filtering cells. Please refer [this doc](https://rasterio.readthedocs.io/en/latest/api/rasterio.mask.html) for details.
        """
        if not hasattr(geo_agent, "geometry"):
            raise TypeError(f"Agent {geo_agent} has no geometry.")
        cells = self.geometric_cells(geo_agent.geometry, **kwargs)
        for cell in cells:
            cell.link_to(agent=geo_agent, link=link)

    def batch_link_by_geometry(
        self, geo_agents: List[Actor], link: Optional[str] = None, **kwargs
    ) -> None:
        """The batch will assign intersecting grids to the main body based on geometry."""
        for geo_agent in geo_agents:
            self.link_by_geometry(geo_agent, link, **kwargs)

    def linked_attr(
        self,
        attr_name: str,
        link: Optional[str] = None,
        nodata: Any = np.nan,
        how: Optional[str] = "only",
    ) -> np.ndarray:
        """Gets the attribute linked to this layer.

        Parameters:
            attr_name:
                The attribute name to search.
            link:
                The link name to get. If None (by default), get the actors located at each cell.
            nodata:
                On the cells without linked agent or located agent, return this nodata.
            how:
                Search mode. Choose the behave when a cell is linked to or have more than one agent.
                    If 'only', raises ABSESpyError when more than one agent is found.
                    If 'random', random choose an agent from all searched agents.

        Returns:
            An array of searched attribute.
        """

        def get_attr(cell: PatchCell, __name):
            return cell.linked_attr(
                attr=__name,
                link=link,
                nodata=nodata,
                how=how,
            )

        return np.vectorize(get_attr)(self.array_cells, attr_name)

    def get_xarray(self, attr_name: str | None = None) -> xr.DataArray:
        """Get the xarray raster layer with spatial coordinates.

        Parameters:
            attr_name:
                The attribute to retrieve. If None (by default), return all available attributes (3D DataArray). Otherwise, 2D DataArray of the chosen attribute.

        Returns:
            Xarray.DataArray data with spatial coordinates of the chosen attribute.
        """
        data = self.get_raster(attr_name=attr_name)
        if attr_name:
            name = attr_name
            data = data.reshape(self.shape2d)
            coords = self.coords
        else:
            coords = {"variable": list(self.attributes)}
            coords |= self.coords
            name = self.name
        return xr.DataArray(
            data=data,
            name=name,
            coords=coords,
        ).rio.write_crs(self.crs)

    def random_positions(
        self,
        k: int = 1,
        where: str | np.ndarray = None,
        prob: str | np.ndarray = None,
        replace: bool = False,
    ) -> List[Coordinate]:
        """
        Choose 'k' `PatchCell` in the layer randomly.

        Parameters:
            k:
                number of patches to choose.
            mask:
                bool mask, only True patches can be choose. If None, all patches are accessible. Defaults to None.
            prob:
                probability of each available position.
            replace:
                If a patch can be chosen more than once. Defaults to False.

        Returns:
            Iterable coordinates of chosen patches.
        """
        where = self._attr_or_array(where).flatten()
        prob = self._attr_or_array(prob).flatten()
        masked_prob = np.where(where, np.nan, prob)
        return norm_choice(
            self.array_cells.flatten(),
            size=k,
            p=masked_prob,
            replace=replace,
        )

    def has_agent(
        self, link: Optional[str] = None, xarray: bool = False
    ) -> np.ndarray:
        """If any actor is linked or existed in each cell.

        Parameters:
            link:
                Link to search. If None (by default), search if any actor is located at in each cell.
            xarray:
                If True, return `Xarray.DataArray` object, `np.ndarray` otherwise (by default).

        Returns:
            A raster data shows weather any actor links or exists.
        """
        if link is None:
            data = np.vectorize(lambda x: x.has_agent)(self.array_cells)
        else:
            data = np.vectorize(lambda x: bool(x.linked(link)))(
                self.array_cells
            )
        if xarray:
            return xr.DataArray(
                data=data,
                coords=self.coords,
                name=link or "has_agent",
            ).rio.write_crs(self.crs)
        return data

    def land_allotment(
        self, agent: Actor, link: str, where: None | str | np.ndarray = None
    ) -> None:
        """
        Allotment of land of this layer to actors.

        Parameters:
            agent:
                the actor to Affirmation of land link.
            link:
                link name of the association.
            where:
                if None, all cells will be selected.
                if str, choose the corresponding attribute raster array as the mask.
                if ndarray, the input array should has the same 2d-shape with this layer.

        Raises:
            ABSESpyError:
                If the shape of input array is not aligned with the shape of this layer.
        """
        mask_ = self._attr_or_array(where)
        cells = self.array_cells[mask_]
        for cell in cells:
            cell.link_to(agent, link)

    def move_agent(self, agent: Actor, position: Coordinate) -> None:
        """Move an Actor to another position of this layer.

        Parameters:
            agent:
                The actor to operate.
            position:
                The position to put on.

        Raises:
            ABSESpyError:
                If the agent is not located at this layer before moving.
        """
        if agent.layer is not self:
            raise ABSESpyError(f"Agent {agent} is not on {self}.")
        agent.put_on_layer(self, position)


class BaseNature(mg.GeoSpace, CompositeModule):
    """The Base Nature Module.
    Note:
        Look at [this tutorial](../features/architectural_elegance.md) to understand the model structure.
        This is NOT a raster layer, but can be seen as a container of different raster layers.
        Users can create new raster layer (i.e., `PatchModule`) by `create_module` method.
        By default, an initialized ABSESpy model will init an instance of this `BaseNature` as `nature` module.

    Attributes:
        major_layer:
            The major layer of nature module. By default, it's the first layer that user created.
        total_bounds:
            The spatial scope of the model's concern. By default, uses the major layer of this model.
    """

    def __init__(self, model, name="nature"):
        CompositeModule.__init__(self, model, name=name)
        crs = self.params.get("crs", CRS)
        mg.GeoSpace.__init__(self, crs=crs)
        self._major_layer = None

    @property
    def major_layer(self) -> PatchModule | None:
        """The major layer of nature module.
        By default, it's the first layer that user created.
        """
        return self._major_layer

    @major_layer.setter
    def major_layer(self, layer: PatchModule):
        if not isinstance(layer, PatchModule):
            raise TypeError(f"{layer} is not PatchModule.")
        self._major_layer = layer
        self.crs = layer.crs

    @property
    def total_bounds(self) -> np.ndarray | None:
        """Total bounds. The spatial scope of the model's concern.
        If None (by default), uses the major layer of this model.
        Usually, the major layer is the first layer sub-module you created.
        """
        if self._total_bounds:
            return self._total_bounds
        if hasattr(self, "major_layer") and self.major_layer:
            return self.major_layer.total_bounds
        return None

    def create_agents_from_gdf(
        self,
        gdf: geopandas.GeoDataFrame,
        unique_id: str = "Index",
        agent_cls: type[Actor] = Actor,
    ) -> ActorsList[Actor]:
        """Create actors from a `geopandas.GeoDataFrame` object.

        Parameters:
            gdf:
                The `geopandas.GeoDataFrame` object to convert.
            unique_id:
                A column name, denotes which column will be converted to unique index of created geo-agents (Social-ecological system Actors).
            agent_cls:
                Agent class to create.

        Returns:
            An `ActorsList` with all new created actors stored.
        """
        creator = mg.AgentCreator(
            model=self.model, agent_class=agent_cls, crs=self.crs
        )
        agents = creator.from_GeoDataFrame(gdf=gdf, unique_id=unique_id)
        self.model.agents.register_a_breed(agent_cls)
        self.model.agents.add(agents)
        return ActorsList(model=self.model, objs=agents)

    def create_module(
        self,
        module_class: Module = PatchModule,
        how: str | None = None,
        **kwargs,
    ) -> PatchModule:
        """Creates a submodule of the raster layer.

        Parameters:
            module_class:
                The custom module class.
            how:
                Class method to call when creating the new sub-module (raster layer). So far, there are three options:
                    `from_resolution`: by selecting shape and resolution.
                    `from_file`: by input of a geo-tiff dataset.
                    `copy_layer`: by copying shape, resolution, bounds, crs, and coordinates of an existing submodule.
                if None (by default), just simply create a sub-module without any custom methods (i.e., use the base class `PatchModule`).
            **kwargs:
                Any other arg passed to the creation method. See corresponding method of your how option from `PatchModule` class methods.

        Returns:
            the created new module.
        """
        module = super().create_module(module_class, how, **kwargs)
        # 如果是第一个创建的模块,则将其作为主要的图层
        if not self.layers:
            self.major_layer = module
        self.add_layer(module)
        return module



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
from typing import Generic, Optional, Tuple, Type, TypeVar

from mesa import Model
from omegaconf import DictConfig

from abses import __version__

from .bases import _Notice
from .container import AgentsContainer
from .human import BaseHuman
from .nature import BaseNature
from .sequences import ActorsList
from .states import States
from .time import TimeDriver

# from .mediator import MainMediator

logger = logging.getLogger(__name__)

# Dynamically load type hints from users' input type
N = TypeVar("N")
H = TypeVar("H")


class MainModel(Generic[N], Model, _Notice, States):
    """
    Base class of a main ABSESpy model.

    Attributes:
        name:
            name of the model. By default, it's the lowercase of class name. E.g.: TestModel -> testmodel.
        settings:
            Structured parameters of the model. Other module or submodules can search the configurations here structurally.
            For an example, if the settings is a nested DictConfig like {'nature': {'test': 3}}, users can access the parameter 'test = 3' by `model.nature.params.test`.
        human:
            The Human module.
        nature:
            The nature module.
        time:
            Time driver.
        params:
            Parameters of the model.
        run_id:
            The run id of the current model. It's useful in batch run.
        agents:
            The container of all agents. One model only has one specific container where all alive agents are stored.
        actors:
            All agents as a list. A model can create multiple lists referring different actors.
    """

    def __init__(
        self,
        parameters: DictConfig = DictConfig({}),
        human_class: Type[H] = BaseHuman,
        nature_class: Type[N] = BaseNature,
        run_id: Optional[int] = None,
        **kwargs,
    ) -> None:
        Model.__init__(self, **kwargs)
        _Notice.__init__(self)
        States.__init__(self)

        self._settings = DictConfig(parameters)
        self._version: str = __version__
        self._human = human_class(self)
        self._nature = nature_class(self)
        self._agents = AgentsContainer(model=self)
        self._time = TimeDriver(model=self)
        self._run_id: int | None = run_id
        self._trigger("initialize", order=("nature", "human"))
        self._trigger("set_state", code=1)  # initial state

    def __repr__(self):
        version = self._version
        return f"<{self.name}-{version}({self.state})>"

    def _trigger(self, _func: str, order: Tuple[str] = None, **kwargs) -> None:
        _obj = {"model": self, "nature": self.nature, "human": self.human}
        if not order:
            order = ("model", "nature", "human")
        for name in order:
            if name not in _obj:
                raise ValueError(f"{name} is not a valid component.")
            getattr(_obj[name], _func)(**kwargs)

    @property
    def run_id(self) -> int | None:
        """The run id of the current model. It's useful in batch run."""
        return self._run_id

    @property
    def name(self) -> str:
        """name of the model. By default, it's the lowercase of class name. E.g.: TestModel -> testmodel."""
        return self.__class__.__name__

    @property
    def version(self) -> str:
        """Report the current version of this model."""
        return self._version

    @property
    def settings(self) -> DictConfig:
        """Structured parameters of the model. Other module or submodules can search the configurations here structurally.

        For an example, if the settings is a nested DictConfig like {'nature': {'test': 3}}, users can access the parameter 'test = 3' by `model.nature.params.test`.
        """
        return self._settings

    @property
    def agents(self) -> AgentsContainer:
        """The container of all agents. One model only has one specific container where all alive agents are stored."""
        return self._agents

    @property
    def actors(self) -> ActorsList:
        """All agents as a list. A model can create multiple lists referring different actors."""
        return self.agents.to_list()

    @property
    def human(self) -> H:
        """The Human class"""
        return self._human

    @property
    def nature(self) -> N:
        """The Nature module"""
        return self._nature

    @property
    def time(self) -> TimeDriver:
        """The time driver & controller"""
        return self._time

    @property
    def params(self) -> DictConfig:
        """The global parameters of this model."""
        return self.settings.get("model", DictConfig({}))

    def run_model(self) -> None:
        """Start running the model, until the end situation is triggered."""
        self._setup()
        while self.running:
            self.step()
            self.time.go()
            self.time.stdout()
        self._end()

    def setup(self):
        """Users can custom what to do when the model is setup and going to start running."""

    def step(self):
        """A step of the model."""

    def end(self):
        """Users can custom what to do when the model is end."""

    def _setup(self):
        self._trigger("setup", order=("model", "nature", "human"))
        self._trigger("set_state", code=2)

    def _end(self):
        self._trigger("end", order=("nature", "human", "model"))
        self._trigger("set_state", code=3)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
import logging
import re
from typing import Any, Iterable

import numpy as np
from scipy import ndimage

logger = logging.getLogger(__name__)


CAMEL_NAME = re.compile(r"(?<!^)(?=[A-Z])")


def get_buffer(
    array: np.ndarray,
    radius: int = 1,
    moor: bool = False,
    annular: bool = False,
) -> np.ndarray:
    """Get a buffer around the array."""
    if radius <= 0:
        raise ValueError(f"Radius must be positive, not {radius}.")
    connectivity = 2 if moor else 1
    struct = ndimage.generate_binary_structure(2, connectivity)
    result = ndimage.binary_dilation(
        array, structure=struct, iterations=radius
    )
    if annular and radius > 1:
        interior = ndimage.binary_dilation(
            array, structure=struct, iterations=radius - 1
        )
        return result & ~interior
    return result


def make_list(element, keep_none=False):
    """Turns element into a list of itself
    if it is not of type list or tuple."""

    if element is None and not keep_none:
        element = []  # Convert none to empty list
    if not isinstance(element, (list, tuple, set, np.ndarray)):
        element = [element]
    elif isinstance(element, (tuple, set)):
        element = list(element)

    return element


def norm_choice(
    a: Iterable[any], size: int = None, p=None, replace=False
) -> Any:
    """
    A more robust random chooser.

    Args:
        a (Iterable[any]): An iterable instance to choose.
        size (int, optional): size of the choice. Defaults to None.
        p (_type_, optional): probability. Defaults to None.
        replace (bool, optional): _description_. Defaults to False.

    Returns:
        Any: _description_
    """
    p = np.array(make_list(p))
    possible_entries = (p > 0).sum()
    if possible_entries == 0:
        p = None
    else:
        p = np.where(p > 0, p, 0.0)
        p /= p.sum()
    if p is None or replace is True:
        return np.random.choice(a, size=size, p=p, replace=replace)
    # 如果有概率的实体 少于要选择的实体，优先返回
    if size is None:
        return np.random.choice(a, p=p, replace=False)
    elif possible_entries < size:
        bounds = a[p > 0]
        rand = np.random.choice(
            a[p == 0], size=(size - possible_entries), replace=False
        )
        selected = np.concatenate([bounds, rand])
    else:
        selected = np.random.choice(a, p=p, size=size, replace=False)
    return selected


def unique_list(*args):
    unique = set()
    for lst in args:
        if not isinstance(lst, Iterable):
            raise TypeError("unique_list can only convert iterable to list")
        unique = unique | set(lst)
    return list(unique)


def iter_func(elements: str) -> callable:
    """
    A decorator broadcasting function to all elements if available.

    elements:
        elements (str): attribute name where object store iterable elements.
        All element in this iterable object will call the decorated function.
    """

    def broadcast(func: callable) -> callable:
        def broadcast_func(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            if not hasattr(self, elements):
                return result
            for element in getattr(self, elements):
                getattr(element, func.__name__)(*args, **kwargs)
            return result

        return broadcast_func

    return broadcast


def camel_to_snake(name):
    """Convert camel name to snake name."""
    # https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    return CAMEL_NAME.sub("_", name).lower()



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from .objects import _BaseObj


class _DynamicVariable:
    """Time dependent variable

    A time dependent function will take the model time driver as
    an input and return its value. The function can also take other
    variables as inputs. The function can be defined as a static
    method of a class or a function.
    """

    def __init__(
        self, name: str, obj: _BaseObj, data: Any, function: Callable
    ) -> None:
        self._name: str = name
        self._obj: _BaseObj = obj
        self._data: Any = data
        self._function: Callable = function

    @property
    def name(self):
        """Get the name of the variable

        Returns
        -------
        name: str
        """
        return self._name

    @property
    def obj(self):
        """Returns a base object instance

        Returns:
            obj:
                _BaseObj
        """
        return self._obj

    @obj.setter
    def obj(self, obj: _BaseObj):
        if not isinstance(obj, _BaseObj):
            raise TypeError("Only accept observer object")
        self._obj = obj

    @property
    def data(self):
        """Returns unused data

        Returns:
            data:
                Any
        """
        return self._data

    @property
    def function(self):
        """Get the function that calculates the variable

        Returns:
            function:
                Callable
        """
        return self._function

    @property
    def time(self):
        """Get the model time driver

        Returns:
            time:
                abses.time._TimeDriver
        """
        return self.obj.time

    def get_required_attributes(self, function: Callable):
        """Get the function required attributes

        Returns:
            required_attributes:
                list[str]
        """
        # Get the source code of the function
        source_code = inspect.getsource(function)
        return [
            attr
            for attr in ["data", "obj", "time", "name"]
            if attr in source_code
        ]

    def now(self) -> Any:
        """Return the dynamic variable function's output

        Returns:
            output:
                Any
        """
        required_attrs = self.get_required_attributes(self.function)
        args = {attr: getattr(self, attr) for attr in required_attrs}
        return self.function(**args)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from typing import Callable, Dict, Set, TypeAlias, Union

from omegaconf import DictConfig

from abses.actor import Actor

from .cells import PatchCell
from .container import AgentsContainer
from .links import LinkContainer
from .modules import CompositeModule, Module
from .sequences import ActorsList, Selection

Actors: TypeAlias = Union[ActorsList, Selection, Actor]
Trigger: TypeAlias = Union[str, Callable]


class HumanModule(Module):
    """The `Human` sub-module base class.

    Note:
        Look at [this tutorial](../features/architectural_elegance.md) to understand the model structure.

    Attributes:
        agents:
            The agents container of this ABSESpy model.
        collections:
            Actor collections defined.
    """

    def __init__(self, model, name=None):
        Module.__init__(self, model, name)
        self._agents = AgentsContainer(model)
        self._collections: Dict[str, Selection] = DictConfig({})

    @property
    def agents(self) -> AgentsContainer:
        """The agents container of this ABSESpy model."""
        return self._agents

    @property
    def collections(self) -> Set[str]:
        """Actor collections defined."""
        return set(self._collections.keys())

    def actors(self, name: str | None = None) -> ActorsList[Actor]:
        """Different selections of agents"""
        if name is None:
            return self.agents.to_list()
        if name not in self._collections:
            raise KeyError(f"{name} is not defined.")
        selection = self._collections[name]
        return self.actors().select(selection)

    def _must_be_actor(self, actor: Actor) -> None:
        if not isinstance(actor, Actor):
            raise TypeError(
                f"Actor must be a subclass of Actor, instead of {type(actor)}."
            )

    def _must_be_cell(self, cell: PatchCell) -> None:
        if not isinstance(cell, PatchCell):
            raise TypeError(
                f"Cell must be a subclass of Cell, instead of {type(cell)}."
            )

    def define(self, name: str, selection: Selection) -> ActorsList:
        """Define a query of actors and save it into collections.

        Parameters:
            name:
                defined name of this group of actors.
            selection:
                Selection query of `Actor`.

        Raises:
            KeyError:
                If the name is already defined.

        Returns:
            The list of actors who are satisfied the query condition.

        Example:
            ```
            # Create 5 actors to query
            model=MainModel()
            model.agents.create(Actor, 5)

            module = HumanModule(model=model)
            actors = module.define(name='first', selection='ids=0')
            >>> len(actors)
            >>> 1

            >>> module.actors('first') == actors
            >>> True
            ```
        """
        if name in self._collections:
            raise KeyError(f"{name} is already defined.")
        selected = self.actors().select(selection)
        self._collections[name] = selection
        return selected


class BaseHuman(CompositeModule, HumanModule, LinkContainer):
    """The Base Human Module.

    Note:
        Look at [this tutorial](../features/architectural_elegance.md) to understand the model structure.
    """

    def __init__(self, model, name="human"):
        LinkContainer.__init__(self)
        HumanModule.__init__(self, model, name)
        CompositeModule.__init__(self, model, name=name)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from loguru import logger

from abses.tools.func import iter_func

from .bases import _Notice
from .objects import _BaseObj
from .states import States

if TYPE_CHECKING:
    from .main import MainModel


class Module(_BaseObj):
    """基本的模块"""

    def __init__(self, model: MainModel, name: Optional[str] = None):
        _BaseObj.__init__(self, model, observer=True, name=name)
        self._open: bool = True

    def __repr__(self) -> str:
        flag = "open" if self.opening else "closed"
        return f"<{self.name}: {flag}>"

    @property
    def opening(self) -> bool:
        """模块处于打开或关闭状态"""
        return self._open

    @iter_func("modules")
    def switch_open_to(self, _open: Optional[bool] = None) -> bool:
        """开启或关闭模块，同时开启或关闭所有子模块"""
        if _open is None:
            return False
        if not isinstance(_open, bool):
            raise TypeError("Accept boolean parameters")
        if self._open is not _open:
            logger.info("%s switch 'open' to %s.", self.name, _open)
            self._open = _open
        return self._open

    # @abstractmethod
    def initialize(self):
        """
        Initialization after handle parameters.
        """

    def setup(self):
        """
        Initialization before handle parameters.
        """

    def step(self):
        """
        每当时间前进时触发
        """

    def end(self):
        """
        每当时间后退时触发
        """


# Composite
class CompositeModule(Module, States, _Notice):
    """基本的组合模块，可以创建次级模块"""

    def __init__(self, model: MainModel, name: str = None) -> None:
        Module.__init__(self, model, name=name)
        States.__init__(self)
        _Notice.__init__(self)
        self._modules: List[Module] = []

    @property
    def modules(self) -> List[Module]:
        """当前模块的次级模块"""
        return self._modules

    def create_module(
        self, module_class: Module, how: Optional[str] = None, **kwargs
    ) -> Module:
        """创建次级模块"""
        if not issubclass(module_class, Module):
            raise TypeError(
                f"Module class {module_class} must inherited from a module."
            )
        if not how:
            module = module_class(model=self._model, **kwargs)
        elif hasattr(module_class, how):
            creating_method = getattr(module_class, how)
            module = creating_method(model=self.model, **kwargs)
        else:
            raise TypeError(f"{how} is not a valid creating method.")
        setattr(self, module.name, module)  # register as module
        self.attach(module)
        self.modules.append(module)  # register as module
        return module

    @iter_func("modules")
    def initialize(self):
        return super().initialize()

    @iter_func("modules")
    def setup(self):
        return super().setup()

    @iter_func("modules")
    def step(self):
        return super().step()

    @iter_func("modules")
    def end(self):
        return super().end()



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
from __future__ import annotations

from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Tuple,
    TypeAlias,
    Union,
)

import mesa_geo as mg
from mesa.space import Coordinate
from shapely import Point

from abses.decision import DecisionFactory
from abses.errors import ABSESpyError
from abses.links import LinkNode
from abses.objects import _BaseObj
from abses.sequences import ActorsList
from abses.tools.func import make_list

# A class that is used to store the position of the agent.


if TYPE_CHECKING:
    from abses.main import MainModel
    from abses.nature import PatchCell


Selection: TypeAlias = Union[str, Iterable[bool]]
Trigger: TypeAlias = Union[Callable, str]


def parsing_string_selection(selection: str) -> Dict[str, Any]:
    """Parses a string selection expression and returns a dictionary of key-value pairs.

    Parameters:
        selection:
            String specifying which breeds to select.

    Returns:
        selection_dict:
            Parsed output as Dictionary
    """
    selection_dict = {}
    if "==" not in selection:
        return {"breed": selection}
    expressions = selection.split(",")
    for exp in expressions:
        left, right = tuple(exp.split("=="))
        selection_dict[left.strip(" ")] = right.strip(" ")
    return selection_dict


def perception_result(name, result, nodata: Any = 0.0) -> Any:
    """clean the result of a perception."""
    if hasattr(result, "__iter__"):
        # raise ABSESpyError('No')
        raise ValueError(
            f"Perception result of '{name}' got type {type(result)} as return."
        )
    return nodata if result is None else result


def perception(
    decorated_func: Callable | None = None, *, nodata: Any | None = None
) -> Callable:
    """Change the decorated function into a perception attribute."""

    def decorator(func) -> Callable:
        @wraps(func)
        def wrapper(self: Actor, *args, **kwargs):
            result = func(self, *args, **kwargs)
            return perception_result(func.__name__, result, nodata=nodata)

        return wrapper

    # 检查是否有参数传递给装饰器，若没有则返回装饰器本身
    return decorator(decorated_func) if decorated_func else decorator


class Actor(mg.GeoAgent, _BaseObj, LinkNode):
    """
    An actor in a social-ecological system (or "Agent" in an agent-based model.)

    Attributes:
        container:
            The container that the actor belongs to.
        layer:
            The layer where the actor is located.
        indices:
            The indices of the cell where the actor is located.
        pos:
            The position of the cell where the actor is located.
        population:
            A list of actors of the same breed as the actor.
        on_earth:
            Whether the actor is standing on a cell.
        here:
            A list of actors that are on the same cell as the actor.

    Methods:
        put_on(self, cell: PatchCell | None = None) -> None
            Places the actor on a cell.
        put_on_layer(self, layer: mg.RasterLayer, pos: Tuple[int, int])
            Specifies a new cell for the actor to be located on.
        selecting(self, selection: Union[str, Dict[str, Any]]) -> bool
            Selects the actor according to specified criteria.
    """

    # when checking the rules
    _freq_levels: Dict[str, int] = {"now": 0, "update": 1, "move": 2, "any": 3}
    __decisions__ = None

    def __init__(
        self,
        model: MainModel,
        observer: bool = True,
        unique_id: Optional[int] = None,
        **kwargs,
    ) -> None:
        _BaseObj.__init__(self, model, observer=observer)
        if not unique_id:
            unique_id = self.model.next_id()
        crs = kwargs.pop("crs", model.nature.crs)
        geometry = kwargs.pop("geometry", None)
        mg.GeoAgent.__init__(
            self, unique_id, model=model, geometry=geometry, crs=crs
        )
        LinkNode.__init__(self)
        self._rules: Dict[str, Dict[str, Any]] = {}
        self._cell: PatchCell = None
        self.container = model.human
        self._decisions = self._setup_decisions()

    def _setup_decisions(self) -> DecisionFactory:
        """Decisions that this actor makes."""
        decisions = make_list(getattr(self, "__decisions__", None))
        return DecisionFactory(self, decisions)

    @property
    def decisions(self) -> DecisionFactory:
        """Decisions that this agent makes."""
        return self._decisions

    # alias of decisions
    d = decisions

    @property
    def layer(self) -> mg.RasterLayer:
        """Get the layer where the agent is located."""
        return None if self._cell is None else self._cell.layer

    @property
    def indices(self) -> Coordinate:
        """Coordinates in the form of (row, col) are used for indexing a matrix, with the origin at the top left corner and increasing downwards."""
        return self._cell.indices

    @property
    def pos(self) -> Coordinate:
        """Coordinates in the form of (x, y) indexed from the bottom left corner."""
        return self._cell.pos

    @pos.setter
    def pos(self, pos) -> None:
        if pos is not None:
            raise AttributeError(f"Set pos by {self.put_on_layer.__name__}")

    @property
    def population(self) -> ActorsList:
        """List of agents of the same breed"""
        return self.model.agents[self.breed]

    @property
    def on_earth(self) -> bool:
        """Whether agent stands on a cell"""
        return bool(self._cell)

    @property
    def here(self) -> ActorsList:
        """Other agents on the same cell as the agent."""
        return self._cell.agents

    def put_on(self, cell: PatchCell | None = None) -> None:
        """Place agent on a cell. If the agent is already located at a cell, it should be located to a cell with the same layer.

        Parameters:
            cell:
                The cell where the agent is to be located. If None (default), remove the subject from the current layer.

        Raises:
            IndexError:
                If the agent is to be moved between different layers.
            TypeError:
                If the agent is to be put on a non-PatchCell object.
        """
        if cell is None:
            # Remove agent
            self._cell = None
            return
        if self.layer and self.layer is not cell.layer:
            raise IndexError(
                f"Trying to move actor between different layers: from {self.layer} to {cell.layer}"
            )
        if not isinstance(cell, mg.Cell):
            raise TypeError(
                f"Actor must be put on a PatchCell, instead of {type(cell)}"
            )
        if self.on_earth:
            self._cell.remove(self)
            self._cell = None
        cell.add(self)
        self._cell = cell
        self.geometry = Point(cell.layer.transform * cell.indices)

    def put_on_layer(
        self, layer: mg.RasterLayer, pos: Tuple[int, int]
    ) -> None:
        """Specifies a new cell for the agent to be located on.

        Parameters:
            layer:
                The layer where the agent is to be located.
            pos:
                The position of the cell where the agent is to be located.

        Raises:
            TypeError:
                If the layer is not a valid RasterLayer type.
        """
        if not isinstance(layer, mg.RasterLayer):
            raise TypeError(f"{layer} is not mg.RasterLayer.")
        cell = layer.cells[pos[0]][pos[1]]
        self.put_on(cell=cell)

    def selecting(self, selection: Union[str, Dict[str, Any]]) -> bool:
        """Either select the agent according to specified criteria.

        Parameters:
            selection:
                Either a string or a dictionary of key-value pairs that represent agent attributes to be checked against.

        Returns:
            Whether the agent is selected or not
        """
        if isinstance(selection, str):
            selection = parsing_string_selection(selection)
        results = []
        for k, v in selection.items():
            attr = getattr(self, k, None)
            if attr is None:
                results.append(False)
            elif attr == v or str(attr) == v:
                results.append(True)
            else:
                results.append(False)
        return all(results)

    def die(self) -> None:
        """Kills the agent (self)"""
        self.model.agents.remove(self)
        for link in self.links:
            self.container.get_graph(link).remove_node(self)
        if self.on_earth:
            self._cell.remove(self)
            del self

    def move_to(self, position: Optional[Coordinate]) -> bool:
        """Move agent to a new position.

        Parameters:
            position:
                The new position to move to.

        Raises:
            ValueError:
                The main body is not on the same layer, please use the put_on method first.
        """
        if not self.layer:
            raise ValueError("Layer is not set.")
        self.layer.move_agent(self, position)

    def loc(self, attribute: str) -> Any:
        """Get attribute data for the cell where the actor is located.

        Parameters:
            attribute : str
                The name of the attribute to get.

        Raises:
            AttributeError:
                If the attribute is not found in the cell where this agent is located.
            ABSESpyError:
                The agent is not in the environment.
        """
        if self.on_earth:
            return self._cell.get_attr(attribute)
        raise ABSESpyError(
            f"You should locate this agent ({self}) to somewhere before get associated attribute."
        )

    def alter_nature(self, attr: str, value: Any) -> None:
        """Alter the nature of the parameters of the cell where the actor is located.

        Parameters:
            attr:
                The name of the parameter to change.
            value:
                The new value to assign to the parameter.

        Raises:
            AttributeError:
                If the attribute is not found in the cell.
        """
        if attr not in self._cell.attributes:
            raise AttributeError(f"Attribute {attr} not found.")
        setattr(self._cell, attr, value)

    def linked(self, link: str) -> ActorsList:
        """Get all other actors linked to this actor.

        Parameters:
            link:
                The link to search for.

        Returns:
            A list of all actors linked to this actor.
        """
        return ActorsList(self.model, super().linked(link))



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

import mesa_geo as mg

from abses import Actor, ActorsList
from abses.errors import ABSESpyError
from abses.links import LinkNode
from abses.sequences import agg_agents_attr

if TYPE_CHECKING:
    from abses.nature import PatchModule


def raster_attribute(func):
    """Turn the method into a property that the patch can extract.
    Examples:
        ```
        class TestCell(Cell):
            @raster_attribute
            def test:
                return 1

        # Using this test cell to create a PatchModule.
        module = PatchModule.from_resolution(
            model=MainModel(),
            shape=(3, 3),
            cell_cls=TestCell,
        )

        # now, the attribute 'test' of TestCell can be accessible in the module, as spatial data (i.e., raster layer).

        >>> module.cell_properties
        >>> set('test')

        >>> array = module.get_raster('test')
        >>> np.array([
            [1, 1, 1],
            [1, 1, 1],
            [1, 1, 1],
        ])
        ```
    """
    func.is_decorated = True
    return property(func)


class PatchCell(mg.Cell, LinkNode):
    """A patch cell of a `RasterLayer`.

    Attributes:
        agents:
            The agents located at here.
        layer:
            The `RasterLayer` where this `PatchCell` belongs.
    """

    def __init__(self, pos=None, indices=None):
        mg.Cell.__init__(self, pos, indices)
        LinkNode.__init__(self)
        self._agents = {}
        self._layer = None

    def __repr__(self) -> str:
        return f"<PatchCell at {self.pos}>"

    @classmethod
    def __attribute_properties__(cls) -> set[str]:
        """Properties that should be found in the `RasterLayer`."""
        return {
            name
            for name, method in cls.__dict__.items()
            if isinstance(method, property)
            and getattr(method.fget, "is_decorated", False)
        }

    @property
    def layer(self) -> PatchModule:
        """`RasterLayer` where this `PatchCell` belongs."""
        return self._layer

    @layer.setter
    def layer(self, layer: PatchModule) -> None:
        if not isinstance(layer, mg.RasterLayer):
            raise TypeError(f"{type(layer)} is not valid layer.")
        self.container = layer.model.human
        self._layer = layer

    @classmethod
    @property
    def breed(cls) -> str:
        """Breed of this `PatchCell`"""
        return cls.__name__

    @property
    def agents(self) -> ActorsList[Actor]:
        """The agents located at here."""
        agents = []
        for _, agents_set in self._agents.items():
            agents.extend(agents_set)
        return ActorsList(self.model, agents)

    def has_agent(self, breed: Optional[str] = None) -> bool:
        """Whether the actor is standing at the current `PatchCell`.

        Parameters:
            breed:
                Specify the breed of agents to search. If None (by default), all breeds of agents are acceptable.

        Returns:
            bool:
                True if there is a qualified principal there, False otherwise.
        """
        return bool(self._agents) if breed is None else breed in self._agents

    def get_attr(self, attr_name: str) -> Any:
        """Gets the value of an attribute or registered property. Automatically update the value if it is the dynamic variable of the layer.

        Parameters:
            attr_name:
                The name of attribute to get.
                
        Returns:
            Any:
                The value of the attribute.
                
        Raises:
            AttributeError:
                Attribute value of the associated patch cell.
        """
        if attr_name in self.layer.dynamic_variables:
            self.layer.dynamic_var(attr_name=attr_name)
        if not hasattr(self, attr_name):
            raise AttributeError(f"{attr_name} not exists in {self.layer}.")
        return getattr(self, attr_name)

    def add(self, agent: Actor) -> None:
        """Adds an `Actor` to here.

        Parameters:
            agent:
                The `Actor` to be added.

        Raises:
            ABSESpyError:
                If the actor is already on earth (at another cell or another layer).
        """
        if agent.on_earth:
            raise ABSESpyError(f"{agent} is already on earth.")
        if not issubclass(agent.__class__, Actor):
            raise TypeError(
                f"The agent to be added should be an instance of 'Actor' or its subclass, not {type(agent)}."
            )
        if agent.breed not in self._agents:
            self._agents[agent.breed] = {agent}
        else:
            self._agents[agent.breed].add(agent)

    def remove(self, agent: Actor) -> None:
        """Removes an Actor here.

        Parameters:
            agent:
                The `Actor` to be removed.

        Raises:
            ABSESpyError:
                If the actor is not on this cell.
        """
        try:
            self._agents[agent.breed].remove(agent)
        except KeyError as err:
            raise ABSESpyError(f"{agent} is not located at {self}.") from err
        agent.put_on()
        if not self._agents[agent.breed]:
            del self._agents[agent.breed]

    def linked(self, link: str) -> ActorsList[Actor]:
        """Gets the body of the link to this patch.

        Parameters:
            link:
                The link type to search.

        Returns:
            An `ActorList` of `Actor` who has association with the patch.

        Raises:
            TypeError:
                The input link should be a string of link name.
            KeyError:
                The searched link is not available in the model.
        """
        if link is None:
            return self.agents
        elif not isinstance(link, str):
            raise TypeError(f"{type(link)} is not valid link name.")
        elif link not in self.links:
            raise KeyError(f"{link} not exists in {self}.")
        else:
            agents = ActorsList(self.model, super().linked(link=link))
        return agents

    def linked_attr(
        self,
        attr: str,
        link: Optional[str] = None,
        nodata: Any = None,
        how: str = "only",
    ) -> Any:
        """Gets the properties of the agent linked to the patch.

        Parameters:
            attr:
                The attribute name to retrieve.
            link:
                The link name to search associations.
            nodata:
                For the agents who don't have such an attribute, return a nodata as a placeholder.
            how:
                Search mode. #TODO

        Returns:
            Any type of retrieved data.
            
        Raises:
            KeyError:
                The searched link is not available in the model.
        """
        try:
            agents = self.linked(link=link)
        except KeyError:
            agents = ActorsList(self.model, [])
        if nodata is None or agents:
            return agg_agents_attr(agents=agents, attr=attr, how=how)
        return nodata

    def get_neighboring_cells(
        self,
        moore: bool = False,
        radius: int = 1,
        include_center: bool = False,
        annular: bool = False,
    ) -> ActorsList:
        """#TODO Get the grid around the patch."""
        cells = self.layer.get_neighboring_cells(
            self.pos, moore=moore, radius=radius, include_center=include_center
        )
        if annular:
            interiors = self.layer.get_neighboring_cells(
                self.pos, moore=moore, radius=radius - 1, include_center=False
            )
            return ActorsList(self.model, set(cells) - set(interiors))
        return ActorsList(self.model, cells)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
ABSESpy - Agent-based Social-ecological System framework in Python
Copyright (c) 2021-2023 Shuang Song

Documentation: https://absespy.github.io/ABSESpy
Examples: https://absespy.github.io/ABSESpy/tutorial/user_guide/
Source: https://github.com/absespy/ABSESpy
"""

__all__ = [
    "__version__",
    "MainModel",
    "BaseHuman",
    "BaseNature",
    "PatchModule",
    "Actor",
    "Decision",
    "DecisionFactory",
    "ActorsList",
    "AgentsContainer",
    "PatchCell",
    "perception",
]
__version__ = "v0.3.1"

from .actor import Actor, perception
from .container import ActorsList, AgentsContainer
from .decision import Decision, DecisionFactory
from .human import BaseHuman
from .main import MainModel
from .nature import BaseNature, PatchCell, PatchModule



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import sys

from loguru import logger

logger.add(
    sys.stderr,
    format="{time} {level} {message}",
    level="WARNING",
    colorize=True,
)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
from collections.abc import Iterable
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Self,
    TypeAlias,
    Union,
    overload,
)

import mesa_geo as mg
import numpy as np

from abses.errors import ABSESpyError

from .tools.func import make_list, norm_choice

if TYPE_CHECKING:
    from .actor import Actor

logger = logging.getLogger("__name__")

Selection: TypeAlias = Union[str, Iterable[bool]]


def get_only_agent(agents: ActorsList) -> Actor:
    """Select one agent"""
    if len(agents) == 0:
        raise ValueError("No agent found.")
    if len(agents) == 1:
        return agents[0]
    raise ValueError("More than one agent.")


def agg_agents_attr(agents: ActorsList, attr, how: str = "only") -> Any:
    """Retrieve the attribute of an either specified or randomly chosen agent."""
    if how == "only":
        return getattr(get_only_agent(agents), attr)
    if how == "random":
        return getattr(np.random.choice(agents), attr)


class ActorsList(list):
    """A list of actors in an agent-based model."""

    def __init__(self, model, objs=()):
        super().__init__(objs)
        self._model = model

    def __repr__(self):
        results = [f"({len(v)}){k}" for k, v in self.to_dict().items()]
        return f"<ActorsList: {'; '.join(results)}>"

    def __getattr__(self, name: str) -> np.ndarray:
        """Return callable list of attributes"""
        # Private variables are looked up normally
        if name[0] == "_":
            return getattr(super(), name)
        if name in self.__dir__():
            return getattr(super(), name)
        return ActorsList.array(self, name)

    def __eq__(self, other: Iterable) -> bool:
        return (
            all(actor in other for actor in self)
            if self._is_same_length(other)
            else False
        )

    @overload
    def __getitem__(self, other: int) -> Actor:
        ...

    @overload
    def __getitem__(self, index: slice) -> Self:
        ...

    def __getitem__(self, index):
        results = super().__getitem__(index)
        return (
            ActorsList(self.model, results)
            if isinstance(index, slice)
            else results
        )

    def _is_same_length(
        self, length: Iterable[Any], rep_error: bool = False
    ) -> bool:
        """Check if the length of input is as same as the number of actors."""
        if not hasattr(self, "__len__"):
            raise ValueError(f"{type(length)} object is not iterable.")
        if len(length) != len(self):
            if rep_error:
                raise ValueError(
                    f"Length of the input {len(length)} mismatch {len(self)} actors."
                )
            return False
        return True

    def to_dict(self) -> Dict[str, Self]:
        """Convert all actors in this list to a dictionary like {breed: ActorList}.

        Returns:
            key is the breed of actors, and values are corresponding actors.
        """
        dic = {}
        for actor in iter(self):
            breed = actor.breed
            if breed not in dic:
                dic[breed] = ActorsList(self.model, [actor])
            else:
                dic[breed].append(actor)
        return dic

    def select(self, selection: Selection) -> Self:
        """
        Returns a new :class:`ActorList` based on `selection`.

        Parameters:
            selection:
                List with same length as the agent list.
                Positions that return True will be selected.
        """
        if isinstance(selection, (str, dict)):
            bool_ = [actor.selecting(selection) for actor in self]
        elif isinstance(selection, (list, tuple, np.ndarray)):
            bool_ = make_list(selection)
        else:
            raise TypeError(f"Invalid selection type {type(selection)}")
        selected = [a for a, s in zip(self, bool_) if s]
        return ActorsList(self.model, selected)

    def ids(self, ids: Iterable[int]) -> List[Actor]:
        """Subsets ActorsList by a `ids`.

        Parameters:
            ids:
                an iterable id list. List[id], ID is an attr of agent obj.

        Returns:
            ActorList: A subset of origin agents list.
        """
        ids = make_list(ids)
        return self.select([agent.id in ids for agent in self])

    def random_choose(
        self,
        size: int = 1,
        prob: Optional[Iterable[float]] = None,
        replace: bool = True,
        as_list: bool = False,
    ) -> Union[Actor, Self]:
        """Randomly choose one or more actors from the current self object.

        Parameters:
            size:
                The number of actors to choose. Defaults to 1.
            prob:
                A list of probabilities for each actor to be chosen.
                If None, all actors have equal probability. Defaults to None.
            replace:
                Whether to sample with replacement. Defaults to True.
            as_list:
                Whether to return the result as a list of actors. Defaults to False.

        Returns:
            An Actor or an ActorList of multiple actors.

        Notes:
            Given the parameter set size=1 and as_list=False, a single Actor object is returned.
            Given the parameter set size>1 and as_list=False, a Self (ActorsList) object is returned.

        Raises:
            ValueError:
                If size is not a positive integer.
        """
        chosen = norm_choice(self, p=prob, size=size, replace=replace)
        if as_list:
            return ActorsList(self.model, objs=chosen)
        if size == 1:
            return chosen[0]
        if size > 1:
            return ActorsList(self.model, chosen)
        raise ValueError(f"Invalid size {size}.")

    def better(
        self, metric: str, than: Optional[Union[Number, Actor]] = None
    ) -> Self:
        """
        Selects the elements of the sequence that are better than a given value or actor
        based on a specified metric.

        Parameters:
            metric:
                The name of the attribute to use as the metric for comparison.
            than:
                The value or actor to compare against. If None, selects the elements with the
                highest value for the specified metric. If a number, selects the elements with
                a value greater than the specified number. If an Actor, selects the elements
                with a value greater than the specified Actor's value for the specified metric.

        Returns:
            A new sequence containing the selected elements.

        Raises:
            ABSESpyError:
                If the `than` parameter is not a Number or an Actor.

        Notes:
            This method compares the values of the specified metric for all elements in the
            sequence and selects the elements that are better than the specified value or actor.
            The comparison is based on the greater than operator (>) for numbers and the
            difference between the values for actors.
        """
        metrics = self.array(attr=metric)
        if than is None:
            return self.select(metrics == max(metrics))
        if isinstance(than, Number):
            return self.select(metrics > than)
        if isinstance(than, mg.GeoAgent):
            diff = self.array(metric) - getattr(than, metric)
            return self.select(diff > 0)
        raise ABSESpyError(f"Invalid than type {type(than)}.")

    def update(self, attr: str, values: Iterable[any]) -> None:
        """Update the specified attribute of each agent in the sequence with the corresponding value in the given iterable.

        Parameters:
            attr:
                The name of the attribute to update.
            values:
                An iterable of values to update the attribute with. Must be the same length as the sequence.

        Raises:
            ValueError:
                If the length of the values iterable does not match the length of the sequence.
        """
        self._is_same_length(values, rep_error=True)
        for agent, val in zip(self, values):
            setattr(agent, attr, val)

    def split(self, where: Iterable[int]) -> np.ndarray:
        """Split agents into N+1 groups.

        Parameters:
            where:
                indexes [size=N] denotes where to split.

        Returns:
            np.ndarray: N+1 groups: agents array
        """
        to_split = np.array(self)
        return np.hsplit(to_split, where)

    def array(self, attr: str) -> np.ndarray:
        """Convert the specified attribute of all actors to a numpy array.

        Parameters:
            attr:
                The name of the attribute to convert to a numpy array.

        Returns:
            A numpy array containing the specified attribute of all actors.
        """
        return np.array([getattr(actor, attr) for actor in self])

    def trigger(self, func_name: str, *args, **kwargs) -> np.ndarray:
        """Call a method with the given name on all actors in the sequence.

        Parameters:
            func_name:
                The name of the method to call on each actor.
            *args:
                Positional arguments to pass to the method.
            **kwargs:
                Keyword arguments to pass to the method.

        Returns:
            An array of the results of calling the method on each actor.
        """
        results = [
            getattr(actor, func_name)(*args, **kwargs) for actor in iter(self)
        ]
        return np.array(results)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
import threading
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Optional,
    Self,
    Tuple,
    Type,
    Union,
)

from abses.actor import Actor
from abses.sequences import ActorsList, Selection
from abses.tools.func import make_list

if TYPE_CHECKING:
    from .main import MainModel

logger = logging.getLogger("__name__")


class AgentsContainer(dict):
    """Singleton AgentsContainer for each model.

    This class is a dictionary-like container for managing agents in a simulation model. It is designed to be a singleton,
    meaning that there is only one instance of this class per model. It provides methods for creating, adding, removing,
    and selecting agents, as well as triggering events.
    """

    _models: Dict[MainModel, AgentsContainer] = {}
    _lock = threading.RLock()

    def __new__(cls: type[Self], model: MainModel) -> Self:
        instance = cls._models.get(model)
        if instance is None:
            instance = super().__new__(cls)
            with cls._lock:
                cls._models[model] = instance
        return instance

    def __init__(self, model: MainModel):
        super().__init__()
        self._model = model
        self._breeds = {}

    def __len__(self):
        return len(self.to_list())

    def __repr__(self):
        # "<ActorsList: >"
        rep = self.to_list().__repr__()[13:-1]
        return f"<AgentsContainer: {rep}>"

    def __getattr__(self, name: str) -> Any:
        if name[0] == "_" or name not in self._breeds:
            return getattr(self, name)
        return self.to_list(name)

    def __contains__(self, name):
        return name in self.to_list()

    @property
    def breeds(self) -> Tuple[str]:
        """Get all breeds in the model"""
        return tuple(self._breeds.keys())

    @property
    def model(self) -> MainModel:
        """The ABSESpy model where the container belongs to."""
        return self._model

    def _register(self, actor: Actor) -> None:
        if actor.breed not in self._breeds:
            self._breeds[actor.breed] = type(actor)
            self[actor.breed] = set()

    def register_a_breed(self, actor_cls: type[Actor]) -> None:
        """Register a new breed of actors in the container.

        Parameters:
            actor_cls:
                The class of the actor to be registered.

        Raises:
            TypeError:
                If the given class is not a subclass of `Actor`.
        """
        if not issubclass(actor_cls, Actor):
            raise TypeError(f"'{actor_cls}' not subclass of 'Actor'.")
        self._breeds[actor_cls.breed] = actor_cls
        self[actor_cls.breed] = set()

    def create(
        self,
        breed_cls: Type[Actor],
        num: int = 1,
        singleton: bool = False,
        **kwargs,
    ) -> Union[Actor, ActorsList[Actor]]:
        """Create one or more actors of the given breed class.

        Parameters:
            breed_cls:
                The breed class of the actor(s) to create.
            num:
                The number of actors to create. Defaults to 1.
            singleton (bool, optional):
                Whether to create a singleton actor. Defaults to False.
            **kwargs:
                Additional keyword arguments to pass to the actor constructor.

        Returns:
            The created actor(s).

        Example:
            ```python
            from abses import Actor, MainModel
            model = MainModel()
            actor = model.agents.create(singleton=True)
            >>> type(actor)
            >>> Actor

            actors = model.agents.create(singleton=False)
            >>> type(actors)
            >>> ActorsList
            ```
        """
        objs = [breed_cls(self._model, **kwargs) for _ in range(num)]
        agents = ActorsList(self._model, objs)
        self.add(agents, register=True)
        if singleton:
            return agents[0] if num == 1 else agents
        return agents

    def to_list(
        self, breeds: Optional[Union[str, Iterable[str]]] = None
    ) -> ActorsList[Actor]:
        """Get all entities of specified breeds to a list.

        Parameters:
            breeds:
                The breed(s) of entities to convert to a list (Optional[Union[str, Iterable[str]]]). If None, all breeds are used.

        Returns:
            ActorsList:
                A list of entities of the specified breeds.
        """
        if breeds is None:
            breeds = self.breeds
        agents = ActorsList(self._model)
        for breed in make_list(breeds):
            agents.extend(self[breed])
        return agents

    def trigger(self, *args, **kwargs) -> Any:
        """Trigger a function for all agents in the container.

        This method calls the `trigger` method of the list of agents in the container,
        passing the same arguments and keyword arguments received by this method.

        Parameters:
            *args:
                Positional arguments to be passed to the `trigger` method of each agent.
            **kwargs:
                Keyword arguments to be passed to the `trigger` method of each agent.

        Returns:
            None
        """
        return self.to_list().trigger(*args, **kwargs)

    def add(
        self,
        agents: Union[Actor, ActorsList, Iterable[Actor]] = None,
        register: bool = False,
    ) -> None:
        """Add one or more actors to the container.

        Parameters:
            agents:
                The actor(s) to add to the container. Defaults to None.
            register:
                Whether to register the actor(s) if they belong to a new breed. Defaults to False.

        Raises:
            TypeError:
                If a breed of the actor(s) is not registered and `register` is False.
        """
        dic = ActorsList(self._model, make_list(agents)).to_dict()
        for k, actors_lst in dic.items():
            if k not in self.breeds:
                if register:
                    self._register(actors_lst[0])
                else:
                    raise TypeError(f"'{k}' not registered.")
            self[k] = self[k].union(actors_lst)

    def remove(self, agent: Actor) -> None:
        """Remove the given agent from the container.

        Parameters:
            agent:
                The agent (actor) to remove.
        """
        self[agent.breed].remove(agent)

    def select(self, selection: Selection) -> ActorsList:
        """Selects the actors that match the given selection criteria.

        Parameters:
            selection:
                The selection criteria to apply.

        Returns:
            A list of actors that match the selection criteria.
        """
        return self.to_list().select(selection)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import logging
import os
from typing import Iterable

from abses.tools.func import make_list

logger = logging.getLogger(__name__)


def get_files_in_folder(
    dir_path,
    suffix: str | Iterable[str] = "",
    iteration: bool = True,
    full_path: bool = False,
) -> list[str]:
    """
    Read all files in a specific path.

    Args:
        dir_path (str): folder path.
        postfix (str|Iterable[str], optional): valid postfix. Defaults to "".
        iteration (bool, optional): if sub-folders include. Defaults to True.
        full_path (bool, optional): return full path of the files, or just the filename. Defaults to False.

    Raises:
        FileNotFoundError: Invalid dir_path.

    Returns:
        list[str]: list of filenames (default),
        or full path of the files (if full_path) is True.
    """
    if not os.path.isdir(dir_path):
        raise FileNotFoundError(f"{dir_path} not a valid folder.")
    paths = []
    valid_postfix = make_list(suffix)
    if iteration:
        for root, _, files in os.walk(dir_path, topdown=False):
            for filename in files:
                validation = [
                    filename.endswith(suffix) for suffix in valid_postfix
                ]
                if not any(validation):
                    continue
                if full_path:
                    path = os.path.join(root, filename)
                    paths.append(path)
                else:
                    paths.append(filename)
    else:
        filename = os.listdir(dir_path)
        paths.append(os.path.join(dir_path, filename))
    return paths
